/*-
 * Copyright (C) @BABOLO  2008 May 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 May 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: getopt.c,v 1.6 2014/04/22 19:28:19 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "parser.h"

#define NOPT  0
#define OPT0  1
#define OPT1  2
/* Multiparameters options are possible with OPTx where x > 1 */
#define ERROUT(A, B) do {ex = (A); errno = (B); goto out;} while (0)

int
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_openopt(BLIN_flag flags, babolo_opt **opt, int argc, char *const *argv, const char *ostr) {/***
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags flags
    babolo_opt *next;
    int ex = EX_OK, j;
    u_char *c;
    int i;

    if  (!ostr) {
        ifBLIN_QX3("+ %08X %"BLIN_X" %d %"BLIN_X" ostr NULL", flags, BLIN_I(opt), argc, BLIN_I(argv));
    } else {
        ifBLIN_QX3( "+ %08X %"BLIN_X" %d %"BLIN_X" ostr=%s~"
                  , flags, BLIN_I(opt), argc, BLIN_I(argv), ostr
                  );
    }
    for (j = 0; !!argv[j]; j++) ifBLIN_QX5(" argv[%d]=%s~", j, argv[j]);
    if  (!opt) {
        ifBLIN_QX0("no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!argv) {
        ifBLIN_QX0("no argv");
        ERROUT(EX_DATAERR, EFAULT);
    }
    next = *opt;
    if  (!((*opt) = malloc(sizeof(babolo_opt)))) {
        ifBLIN_QX0("no mem #0");
        ERROUT(EX_OSERR, ENOMEM);
    }
    (*opt)->next = next;
    (*opt)->flags = flags & ~(Bpars_NOAL | Bpars_NPRM | Bpars_EOOP);
    (*opt)->argc = argc;
    (*opt)->argv = (u_char**)argv;
    (*opt)->ind = (flags & Bpars_NONU) ? 1 : 0;
    (*opt)->plc = (u_char*)"";
    (*opt)->arg = NULL;
    if  (!!ostr) {
        if  (!((*opt)->opts = calloc(1, 256))) {
            free(*opt);
            *opt = next;
            ifBLIN_QX0("no mem #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        for (i = 0; ostr[i];) {
            c = &(*opt)->opts[((u_char*)ostr)[i++]];
            if  (ostr[i] == ':') {
                *c = OPT1;
                i++;
            } else {
                *c = OPT0;
        }   }
    } else if (!!next) {
        (*opt)->opts = next->opts;
        (*opt)->flags |= Bpars_NOAL;
    } else {
        free(*opt);
        *opt = next;
        ifBLIN_QX0("ostr and next both NULLs");
        ERROUT(EX_DATAERR, EINVAL);
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_closeopt(babolo_opt **opt) {                                                              /****
 *****************************************************************************************************
 *****************************************************************************************************/
    BLIN_flag flags = blin_internal_flags & BLIN_MASK;
#   define blin_internal_flags flags
    babolo_opt *next;
    int ex = 0;

    if  (!opt) {
        ifBLIN_QX0("no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!!*opt) flags = (*opt)->flags;
    while (!!*opt) {
        ifBLIN_QX3("+ %"BLIN_X" %"BLIN_X, BLIN_I(opt), BLIN_I(*opt));
        next = (*opt)->next;
        if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
        free(*opt);
        *opt = next;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_getopt(babolo_opt **opt) {                                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    BLIN_flag flags = blin_internal_flags & BLIN_MASK;
#   define blin_internal_flags flags
    babolo_opt *next;
    int ex = 0;

    if  (!opt) {
        ifBLIN_QX0("no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!*opt) goto out;
    flags = (*opt)->flags;
in:
    ifBLIN_QX3( "+ %"BLIN_X" %"BLIN_X" ind=%d argc=%d"
              , BLIN_I(opt), BLIN_I(*opt), (*opt)->ind, (*opt)->argc
              );
    ex = 0;
    if  ((*opt)->flags & Bpars_EOOP) {
        goto out;
    } else if (*((*opt)->plc) == 0) {
        ifBLIN_QX4(" ++ %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
        if  (!(*opt)->argv) {
            ifBLIN_QX0("no argv");
            ERROUT(EX_DATAERR, EFAULT);
        }
        if  ((*opt)->ind >= (*opt)->argc) {
            /* Argument is absent */
            next = (*opt)->next;
            if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
            free(*opt);
            *opt = next;
            if  (!*opt) goto out;
            goto in;
        }
        if  (!((*opt)->plc = (*opt)->argv[(*opt)->ind++])) {
            ifBLIN_QX0("no argv[%d]", (*opt)->ind);
            ERROUT(EX_DATAERR, EFAULT);
        }
        if  (*((*opt)->plc) != '-') {
            /* Argument is not an option */
            (*opt)->flags |= Bpars_EOOP;
            goto out;
        }
        ex = *(++(*opt)->plc);
        if  (ex == '-' && *(++(*opt)->plc) == 0) {
            /* "--" => end of options */
            ex = 0;
            goto out;
        }
        ++(*opt)->plc;
        if  (ex == 0) {
            /* Solitary '-', treat as a '-' option */
            ex = '-';
        }
    } else {
        ifBLIN_QX4(" @@ %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
        ex = *((*opt)->plc++);
    }
    ifBLIN_QX4(" getopt %02X=%c", ex, ex);
    (*opt)->flags &= ~Bpars_NPRM;
    switch((*opt)->opts[ex]) {
    case NOPT:
        ifBLIN_QX4(" NOPT %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
        if  (!((*opt)->flags & Bpars_ESYM)) {
            ifBLIN_QX0("invalid option");
            ERROUT(-1, EINVAL);
        }
        break;
    case OPT0:
        ifBLIN_QX4(" OPT0 %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
        break;
    default:
        setBpars_NPRM((*opt)->flags, 1);
mid:
        ifBLIN_QX4(" OPT1 %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
        if  (*((*opt)->plc) != 0) {
            (*opt)->arg = (*opt)->plc;
        } else if ((*opt)->ind < (*opt)->argc) {
            (*opt)->arg = (*opt)->argv[(*opt)->ind++];
        } else {
            /* option-argument absent */
            u_int32_t f;

            next = (*opt)->next;
            if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
            f = (*opt)->flags;
            free(*opt);
            *opt = next;
            if  (!*opt) {
                if  (!(f & Bpars_EPRM)) {
                    ifBLIN_QX0("no argument for -%c", ex);
                    ERROUT(-1, ENOENT);
                }
                goto out;
            }
            goto mid;
        }
        (*opt)->plc = (u_char*)"";
    }
    ifBLIN_QX4(" -- %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

char *
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_getoptarg(babolo_opt **opt, int argn) {                                                   /****
 *****************************************************************************************************
 *****************************************************************************************************/
    BLIN_flag flags = blin_internal_flags & BLIN_MASK;
#   define blin_internal_flags flags
    int ex = 0;
    char *c;

    c = NULL;
    if  (!opt) {
        ifBLIN_QX0("no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!*opt) goto out;
    flags = (*opt)->flags;
    ifBLIN_QX3("+ %"BLIN_X" %"BLIN_X, BLIN_I(opt), BLIN_I(*opt));
    if  (argn > 0) {
        ifBLIN_QX0("multiargument flags not realised");
    } else if ((flags & Bpars_NPRM) > argn) {
        c = (char*)((*opt)->arg);
    }
    if  (!!c) {
        ifBLIN_QX3("- =%s~", c);
    } else {
        ifBLIN_QX3("-");
    }
out:
    return(c);
#   undef blin_internal_flags
}

char *
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_getarg(babolo_opt **opt) {                                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    BLIN_flag flags = blin_internal_flags & BLIN_MASK;
#   define blin_internal_flags flags
    babolo_opt *next;
    int ex = 0;
    char *c;

    c = NULL;
    if  (!opt) {
        ifBLIN_QX0("no opt");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!*opt) goto out;
    flags = (*opt)->flags;
in:
    ifBLIN_QX3( "+ %"BLIN_X" %"BLIN_X" ind=%d argc=%d"
              , BLIN_I(opt), BLIN_I(*opt), (*opt)->ind, (*opt)->argc
              );
    if  (*((*opt)->plc) == 0) {
        ifBLIN_QX4(" ++ %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
        if  (!(*opt)->argv) {
            ifBLIN_QX0("no argv");
            ERROUT(EX_DATAERR, EFAULT);
        }
        if  ((*opt)->ind >= (*opt)->argc) {
            /* Argument is absent */
            next = (*opt)->next;
            if  (!((*opt)->flags & Bpars_NOAL)) free((*opt)->opts);
            free(*opt);
            *opt = next;
            if  (!*opt) goto out;
            goto in;
        }
        if  (!((*opt)->plc = (*opt)->argv[(*opt)->ind++])) {
            ifBLIN_QX0("no argv[%d]", (*opt)->ind);
            ERROUT(EX_DATAERR, EFAULT);
    }   }
    c = (char*)((*opt)->plc);
    (*opt)->plc = (u_char*)"";
    ifBLIN_QX4(" -- %"BLIN_X" [%d]%s~", BLIN_I(*opt), (*opt)->ind, (*opt)->plc);
out:
    ifBLIN_QX3("- =%s~", c);
    return(c);
#   undef blin_internal_flags
}

static const char topr[256] =
".0123456789XXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX"
"XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX"
"XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX"
"XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX" "XXXXXXXXXXXXXXXX"
;
void
/*****************************************************************************************************
 *****************************************************************************************************/
babolo_dumpopt(babolo_opt *opt) {                                                                /****
 *****************************************************************************************************
 *****************************************************************************************************/
    BLIN_flag flags = blin_internal_flags & BLIN_MASK;
#   define blin_internal_flags flags
    int ex = 0, i, j;

    for (i = 0; !!opt; i++, opt = opt->next) {
        flags = opt->flags;
        fprintf( stderr, "%3d %"BLIN_X" flags=%08X argc=%d ind=%d\n"
               , i, BLIN_I(opt), opt->flags, opt->argc, opt->ind
               );
        if  (!!opt->opts) {
            fprintf(stderr, "opts 0123 4567 89AB CDEF\n");
            for (j = 0; j < 16; j++) {
                fprintf( stderr, " %02X: %c%c%c%c %c%c%c%c %c%c%c%c %c%c%c%c\n"
                       , j * 16
                       , topr[opt->opts[j * 16 +  0]], topr[opt->opts[j * 16 +  1]]
                       , topr[opt->opts[j * 16 +  2]], topr[opt->opts[j * 16 +  3]]
                       , topr[opt->opts[j * 16 +  4]], topr[opt->opts[j * 16 +  5]]
                       , topr[opt->opts[j * 16 +  6]], topr[opt->opts[j * 16 +  7]]
                       , topr[opt->opts[j * 16 +  8]], topr[opt->opts[j * 16 +  9]]
                       , topr[opt->opts[j * 16 + 10]], topr[opt->opts[j * 16 + 11]]
                       , topr[opt->opts[j * 16 + 12]], topr[opt->opts[j * 16 + 13]]
                       , topr[opt->opts[j * 16 + 14]], topr[opt->opts[j * 16 + 15]]
                       );
        }   }
        if  (!!opt->arg) fprintf(stderr, "arg=%s~\n", opt->arg);
        if  (!!opt->plc) fprintf(stderr, "plc=%s~\n", opt->plc);
        if  (!!opt->argv) {
            for (j = 0; !!opt->argv[j]; j++) {
                fprintf(stderr, "%cargv[%d]=%s~\n", (j == opt->ind) ? '>' : ' ', j, opt->argv[j]);
    }   }   }
    ifBLIN_QX3("- %d", ex);
#   undef blin_internal_flags
}
